# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2
from . import job_pb2 as job__pb2
from . import nbox_ws_pb2 as nbox__ws__pb2


class WSJobServiceStub(object):
    """
    WSJobService is used by nbox to interface with Web Server. WSJobService manages
    DB and NBX Platform Information. All compute resource related actions are fowarded
    to WSJobsService running on AWS.

    // Generated according to https://cloud.google.com/apis/design/standard_methods
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetJob = channel.unary_unary(
                '/jobs.ws.WSJobService/GetJob',
                request_serializer=nbox__ws__pb2.JobInfo.SerializeToString,
                response_deserializer=job__pb2.Job.FromString,
                )
        self.ListJobs = channel.unary_unary(
                '/jobs.ws.WSJobService/ListJobs',
                request_serializer=nbox__ws__pb2.ListJobsRequest.SerializeToString,
                response_deserializer=nbox__ws__pb2.ListJobsResponse.FromString,
                )
        self.UploadJobCode = channel.unary_unary(
                '/jobs.ws.WSJobService/UploadJobCode',
                request_serializer=nbox__ws__pb2.UploadCodeRequest.SerializeToString,
                response_deserializer=job__pb2.Job.FromString,
                )
        self.CreateJob = channel.unary_unary(
                '/jobs.ws.WSJobService/CreateJob',
                request_serializer=nbox__ws__pb2.CreateJobRequest.SerializeToString,
                response_deserializer=job__pb2.Job.FromString,
                )
        self.UpdateJob = channel.unary_unary(
                '/jobs.ws.WSJobService/UpdateJob',
                request_serializer=nbox__ws__pb2.UpdateJobRequest.SerializeToString,
                response_deserializer=job__pb2.Job.FromString,
                )
        self.DeleteJob = channel.unary_unary(
                '/jobs.ws.WSJobService/DeleteJob',
                request_serializer=nbox__ws__pb2.JobInfo.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.GetJobLogs = channel.unary_stream(
                '/jobs.ws.WSJobService/GetJobLogs',
                request_serializer=nbox__ws__pb2.JobLogsRequest.SerializeToString,
                response_deserializer=nbox__ws__pb2.JobLog.FromString,
                )
        self.TriggerJob = channel.unary_unary(
                '/jobs.ws.WSJobService/TriggerJob',
                request_serializer=nbox__ws__pb2.JobInfo.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_wrappers__pb2.BoolValue.FromString,
                )
        self.UpdateRun = channel.unary_unary(
                '/jobs.ws.WSJobService/UpdateRun',
                request_serializer=nbox__ws__pb2.UpdateRunRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.GetRunToken = channel.unary_unary(
                '/jobs.ws.WSJobService/GetRunToken',
                request_serializer=nbox__ws__pb2.JobInfo.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.FromString,
                )


class WSJobServiceServicer(object):
    """
    WSJobService is used by nbox to interface with Web Server. WSJobService manages
    DB and NBX Platform Information. All compute resource related actions are fowarded
    to WSJobsService running on AWS.

    // Generated according to https://cloud.google.com/apis/design/standard_methods
    """

    def GetJob(self, request, context):
        """GetJob returns the job with the given info
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListJobs(self, request, context):
        """ListJobs returns a list of jobs for the given workspace/user
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UploadJobCode(self, request, context):
        """To upload Job Code to S3
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateJob(self, request, context):
        """CreateJob creates a new job
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateJob(self, request, context):
        """UpdateJob updates the given job
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteJob(self, request, context):
        """DeleteJob deletes the given job
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetJobLogs(self, request, context):
        """GetJobLogs returns logs from given job
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TriggerJob(self, request, context):
        """TriggerJob triggers the given job
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateRun(self, request, context):
        """UpdateRun update the run for given job
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetRunToken(self, request, context):
        """GetRunToken returns a tokn to be used for `/UpdateRun`
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_WSJobServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetJob': grpc.unary_unary_rpc_method_handler(
                    servicer.GetJob,
                    request_deserializer=nbox__ws__pb2.JobInfo.FromString,
                    response_serializer=job__pb2.Job.SerializeToString,
            ),
            'ListJobs': grpc.unary_unary_rpc_method_handler(
                    servicer.ListJobs,
                    request_deserializer=nbox__ws__pb2.ListJobsRequest.FromString,
                    response_serializer=nbox__ws__pb2.ListJobsResponse.SerializeToString,
            ),
            'UploadJobCode': grpc.unary_unary_rpc_method_handler(
                    servicer.UploadJobCode,
                    request_deserializer=nbox__ws__pb2.UploadCodeRequest.FromString,
                    response_serializer=job__pb2.Job.SerializeToString,
            ),
            'CreateJob': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateJob,
                    request_deserializer=nbox__ws__pb2.CreateJobRequest.FromString,
                    response_serializer=job__pb2.Job.SerializeToString,
            ),
            'UpdateJob': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateJob,
                    request_deserializer=nbox__ws__pb2.UpdateJobRequest.FromString,
                    response_serializer=job__pb2.Job.SerializeToString,
            ),
            'DeleteJob': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteJob,
                    request_deserializer=nbox__ws__pb2.JobInfo.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'GetJobLogs': grpc.unary_stream_rpc_method_handler(
                    servicer.GetJobLogs,
                    request_deserializer=nbox__ws__pb2.JobLogsRequest.FromString,
                    response_serializer=nbox__ws__pb2.JobLog.SerializeToString,
            ),
            'TriggerJob': grpc.unary_unary_rpc_method_handler(
                    servicer.TriggerJob,
                    request_deserializer=nbox__ws__pb2.JobInfo.FromString,
                    response_serializer=google_dot_protobuf_dot_wrappers__pb2.BoolValue.SerializeToString,
            ),
            'UpdateRun': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateRun,
                    request_deserializer=nbox__ws__pb2.UpdateRunRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'GetRunToken': grpc.unary_unary_rpc_method_handler(
                    servicer.GetRunToken,
                    request_deserializer=nbox__ws__pb2.JobInfo.FromString,
                    response_serializer=google_dot_protobuf_dot_wrappers__pb2.StringValue.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'jobs.ws.WSJobService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class WSJobService(object):
    """
    WSJobService is used by nbox to interface with Web Server. WSJobService manages
    DB and NBX Platform Information. All compute resource related actions are fowarded
    to WSJobsService running on AWS.

    // Generated according to https://cloud.google.com/apis/design/standard_methods
    """

    @staticmethod
    def GetJob(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/jobs.ws.WSJobService/GetJob',
            nbox__ws__pb2.JobInfo.SerializeToString,
            job__pb2.Job.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListJobs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/jobs.ws.WSJobService/ListJobs',
            nbox__ws__pb2.ListJobsRequest.SerializeToString,
            nbox__ws__pb2.ListJobsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UploadJobCode(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/jobs.ws.WSJobService/UploadJobCode',
            nbox__ws__pb2.UploadCodeRequest.SerializeToString,
            job__pb2.Job.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateJob(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/jobs.ws.WSJobService/CreateJob',
            nbox__ws__pb2.CreateJobRequest.SerializeToString,
            job__pb2.Job.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UpdateJob(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/jobs.ws.WSJobService/UpdateJob',
            nbox__ws__pb2.UpdateJobRequest.SerializeToString,
            job__pb2.Job.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeleteJob(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/jobs.ws.WSJobService/DeleteJob',
            nbox__ws__pb2.JobInfo.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetJobLogs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/jobs.ws.WSJobService/GetJobLogs',
            nbox__ws__pb2.JobLogsRequest.SerializeToString,
            nbox__ws__pb2.JobLog.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def TriggerJob(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/jobs.ws.WSJobService/TriggerJob',
            nbox__ws__pb2.JobInfo.SerializeToString,
            google_dot_protobuf_dot_wrappers__pb2.BoolValue.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UpdateRun(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/jobs.ws.WSJobService/UpdateRun',
            nbox__ws__pb2.UpdateRunRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetRunToken(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/jobs.ws.WSJobService/GetRunToken',
            nbox__ws__pb2.JobInfo.SerializeToString,
            google_dot_protobuf_dot_wrappers__pb2.StringValue.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



"""
Jupyter Notebooks are user by millions of developers world wide, so we said why not just
create an operator that runs the jupyter notebook and creates a python file out of it.
Catch: Currently we do not support notebooks that have a ``!`` for shell commands, please
replace them with ``subprocess.call(shelex.split(<com>))``

.. code-block:: python

  from nbox.lib.notebook import NotebookRunner
  op = NotebookRunner("./sample.ipynb")
  op() # call and it will do the rest
"""

import sys
import ast
import json
import importlib.util
from functools import lru_cache

import nbox.utils as U
from nbox import operator, Operator, logger
from nbox.lib.shell import ShellCommand

@lru_cache(maxsize=1)
def get_py_builtin_all():
  # a cheap 99% solution that works for 99% of people
  builtin_path = U.join(U.folder(U.folder(__file__)), "assets/builtin")
  with open(builtin_path, "r") as f:
    items = []
    for l in f:
      if not l.startswith("#"):
        items.append(l.strip())
  return items

def get_package_version(name = 'tensorflow'):
  # https://stackoverflow.com/questions/1051254/check-if-python-package-is-installed

  # I don't know how to remove some of the packages which are not part of builtins
  # but are still in the python lib like json or logging.

  if name in sys.builtin_module_names or name in get_py_builtin_all():
    return ""

  if name in sys.modules:
    version = getattr(sys.modules[name], "__version__", None)
    if version:
      return version
    return ""
  spec = importlib.util.find_spec(name)
  if spec is not None:
    # If you choose to perform the actual import ...
    module = importlib.util.module_from_spec(spec)
    sys.modules[name] = module
    spec.loader.exec_module(module)
    version = getattr(sys.modules[name], "__version__", None)
    del(sys.modules[name])
    if version:
      return version
    return ""

  # TODO: when can't find a package -> maybe it is a file relative to the notebook
  return False

class ImportNodeVisitor(ast.NodeVisitor):
  def __init__(self):
    self.import_modules = set()

  def visit_Import(self, node: ast.Import):
    for n in node.names:
      n = n.name.split(".")[0]
      self.import_modules.add(n)

  def visit_ImportFrom(self, node: ast.ImportFrom):
    self.import_modules.add(node.module)


def process_codeblocks(cells) -> str:
  """This function is responsible for unpacking the code blocks and creating the python file from it."""
  code_lines = []
  for cell in cells:
    if not cell["cell_type"] == "code":
      # this is not even code, probably markdown, so skip
      # ideally I would like to write down everything, but
      # it's okay here because that doesn't matter to computer
      # maybe an AI can read it and do something.
      continue
    code_lines.extend(cell["source"])
    code_lines.append("\n")

  # let's make this shit tricky:
  # Jupyter has support for built in shell commands by adding a ! like this
  #     !echo "hello"
  # moreover it has support for this in loops and all with reference to actual
  # python values by adding a $ in front of the variable name
  #     for x in range(2):
  #       !echo "alpha-$x"
  # since things like this cannot be really parsed by python ast, we need to
  # perform a modifications to the code before we parse it
  # remember this can also be in multiple lines by adding \ at the end
  #
  # this is a very hard problem to solve because of the number of edge cases
  # if there is a ! inside triple quotes, then it cannot be updated
  # however it is very hard to determine the triple quoted strings without
  # running the code. the chicken and egg problem
  # so for now we are saying if you have that kindly remove it.

  # jupyter notebook also has support for magic commands through % and %% like
  # %matplotlib inline
  # %%time
  # so we need to remove those as well, the good thing is that they cannot persist
  # across multiple lines, so we can just remove them line by line
  for i, line in enumerate(code_lines):
    if line.startswith("%"):
      line = "# " + line
      code_lines[i] = line
  full_code_raw = "".join(code_lines)

  # go over all the imports and sort them out
  nv = ImportNodeVisitor()
  try:
    nv.visit(ast.parse(full_code_raw))
  except Exception as e:
    # str(e) => (<unknown>, line 187) # get the 187
    eline = int(str(e)[:-1].split("line")[1].strip())
    line = code_lines[eline-1]
    if line.strip().startswith("!"):
      # if the line is commented out, then it's okay
      logger.error("Error:")
      logger.error(" reason: this is supported by jupyter notebook but not python")
      logger.error("    fix: subprocess.call(shelex.split(<com>))")
      logger.error(f"    >>> {line}")
      raise
    else:
      raise e
  import_modules = nv.import_modules
  module_to_version = {}
  for x in import_modules:
    v = get_package_version(x)
    if v:
      module_to_version[x] = v

  # there is no need to brong the imports to top because scoping will be unpredictable
  return full_code_raw, module_to_version

@operator()
def nb2script(nb, path: str = None):
  """This function takes in a notebook file and creates a python file from it.
  
  Args:
    nb (str): path to the notebook file
    path (str): path to the python file to be created, if `None` same name as notebook
  """
  # read the file and process it to get the code
  with open(nb, "r") as f:
    code = json.load(f)
  full_code_raw, import_modules = process_codeblocks(code["cells"])
  if not full_code_raw:
    raise ValueError("Oops! No code blocks found in the notebook")

  # Start populating the header metadata and the complete the file
  kernel = "N/A"
  nb_version = f"{code['nbformat']}.{code['nbformat_minor']}"
  meta = code["metadata"]
  if meta:
    if "kernelspec" in meta:
      kernel = meta["kernelspec"]["display_name"]

  import_string = "\n".join(["# " + x for x in json.dumps(import_modules, indent=2).splitlines()])
  _metadata = f'''
# -------------
# Here's some information on the notebook:
#   NB Version: {nb_version}
#         Path: {nb}
#       Kernel: {kernel}
#
# Here's a list of found imports:
{import_string}
# -------------
  '''.strip()

  py_file = f'''
# Auto-generated by Nb2Script Operator: Do Not Modify
{_metadata}

# START >
{full_code_raw}
# < END
'''

  # write the file to correct location
  if path is None:
    path = nb.replace(".ipynb", ".py")
  with open(path, "w") as f:
    f.write(py_file)

  return import_modules


def update_requirements(req_path, mod_versions):
  lines = []
  mods_added = set()
  with open(req_path, "r") as f:
    for line in f:
      if line.startswith("#"):
        lines.append(line)
      else:
        mod = line.split("==")[0]
        if mod in mod_versions:
          ver = mod_versions[mod]
          line = f"\n{mod}"
          if ver:
            line += f"=={ver}"
          lines.append(line + "\n")
          mods_added.add(mod)
        else:
          lines.append(line)
  for mod, ver in mod_versions.items():
    if mod not in mods_added:
      ver = mod_versions[mod]
      line = f"\n{mod}"
      if ver:
        line += f"=={ver}"
      lines.append(line + "\n")
  with open(req_path, "w") as f:
    f.write("".join(lines))


class NotebookRunner(Operator):
  def __init__(self, nb: str, outfile: str = "out.py"):
    """Run any notebook as a python file.
    
    Args:
      nb (str): path to the notebook file
      outfile (str): path to the python file to be created
    """
    super().__init__()
    self.nb = nb

    # create the target runner file
    imports = nb2script(self.nb, outfile)

    # now update the imports of this file
    update_requirements("./requirements.txt", imports)

    # define the shell command to run this thing
    self.run = ShellCommand(f"/job/venv/bin/python3 {outfile}")

  def forward(self):
    self.run()



"""
This is CLI for ``nbox``, it is meant to be as simple to use as possible. We using ``python-fire`` for building
our CLI, which means you can access underlying system with same commands as you use with python. For maximum
convinience we recommend adding an alias for ``nbx`` as follows:

.. code-block:: bash
  
  [zshrc]
  echo "\\nalias nbx='python3 -m nbox'\\n" >> ~/.zshrc
  source ~/.zshrc

  [bashrc]
  echo "\\nalias nbx='python3 -m nbox'\\n" >> ~/.bashrc
  source ~/.bashrc


SSH into Instances
------------------

You can SSH into your instance with the ``nbx tunnel`` command. This command will create a tunnel
to your instance and start an SSH session.

.. code-block:: bash

  nbx tunnel 8000 -i "instance-name"

Documentation
-------------
"""

import os
import sys
import fire
from json import dumps

import nbox.utils as U
from nbox.jobs import Job, Serve
from nbox.init import nbox_ws_v1
from nbox.auth import init_secret, ConfigString
from nbox.instance import Instance
from nbox.sub_utils.ssh import tunnel
from nbox.relics import RelicsNBX
from nbox.version import __version__ as V

logger = U.get_logger()

def global_config(workspace_id: str = ""):
  """Set global config for ``nbox``"""
  secret = init_secret()
  if workspace_id:
    secret.put(ConfigString.workspace_id.value, workspace_id, True)
    logger.info(f"Global Workspace ID set to {workspace_id}")


def open_home():
  """Open current NBX platform"""
  from .auth import secret
  import webbrowser
  webbrowser.open(secret.get("nbx_url"))


def get(api_end: str, **kwargs):
  """Get any command 

  Args:
    api_end (str): something like '/v1/api'
  """
  api_end = api_end.strip("/")
  if nbox_ws_v1 == None:
    raise RuntimeError("Not connected to NimbleBox.ai webserver")
  out = nbox_ws_v1
  for k in api_end.split("/"):
    out = getattr(out, k)
  res = out(_method = "get", **kwargs)
  sys.stdout.write(dumps(res))
  sys.stdout.flush()

def login():
  """Re authenticate ``nbox``. NOTE: Will remove all added keys to ``secrets``"""
  fp = U.join(U.env.NBOX_HOME_DIR(), "secrets.json")
  os.remove(fp)
  init_secret()

def version():
  logger.info("NimbleBox.ai Client Library")
  logger.info(f"    nbox version: {V}")


def why():
  print("\nWhy we build NimbleBox?\n")
  print("  * Artificial intelligence will be the most important technology of the 21st century.")
  print("  * Every single piece of software ever written will need to be upgraded.")
  print("  * Energy spent per code token will increase exponentially to handle the bandwidth of AI.")
  print("  * AI is still software and software engineering is hard.")
  print("  * Nimblebox is a general purpose tool to build and manage such operations.")
  print("\nIf you like what we are building, come work with us.\n\nWith Love,\nNimbleBox.ai\n")


def main():
  fire.Fire({
    "build"   : Instance,
    "config"  : global_config,
    "get"     : get,
    "jobs"    : Job,
    "login"   : login,
    "open"    : open_home,
    "relics"  : RelicsNBX,
    "serve"   : Serve,
    "tunnel"  : tunnel,
    "version" : version,
    "why"     : why,
  })

if __name__ == "__main__":
  main()



import os
import cloudpickle
from glob import glob
from uuid import uuid4
from pathlib import Path
from shutil import rmtree
from hashlib import sha256
from typing import Any, Union, List

from nbox.utils import logger, FileLogger, env
from nbox.relics.base import BaseStore

class RelicLocal(BaseStore):
  """
  The Relic is a part of a filesystem, however `RelicLocal` is an exception since the data that it recieves might
  optionally be a python object in which case it needs to store that.

  Cache structure is like this:

  {cache_dir}/
    _objects.bin # contains the entirity of information on this cache
    activity.log # contains the logs of this cache
    items/
      fdedc958b417adf63278938efa53c3b381f576446aede80bbc8f2c05320fcb4b
      1459d663d14b7b7ad82ebe5c98c8a0397b21d4e2b9f4711562746a5cb48f86c4
      ...
  """
  def __init__(self, relic_name: str, workspace_id: str, create: bool = False):
    self.relic_name = relic_name
    self.workspace_id = workspace_id
    self.cache_dir = os.path.join(env.NBOX_HOME_DIR(), "relics")
    logger.info(f"Connecting object store: {self.cache_dir}")
    self._objects = {} # <key: item_path>
    self._objects_bin_path = f"{self.cache_dir}/_objects.bin"
    self._file_logger_path = f"{self.cache_dir}/activity.log"

    # Create the cache directory if it doesn't exist
    if not os.path.exists(self.cache_dir):
      Path(self.cache_dir).mkdir(parents=create)
    
    # load the data from the cache
    if os.path.isdir(self.cache_dir):
      if not os.path.exists(self._objects_bin_path):
        self._objects = {}
      else:
        with open(self._objects_bin_path, "rb") as f:
          self._objects = cloudpickle.load(f)
      if not os.path.exists(f"{self.cache_dir}/items"):
        os.makedirs(f"{self.cache_dir}/items")
      self._logs = FileLogger(self._file_logger_path)
    else:
      raise Exception(f"{self.cache_dir} is not a directory")

  def get_id(self, key: str = "") -> str:
    _key = sha256(key.encode('utf-8')).hexdigest()
    item_path = f"{self.cache_dir}/items/{_key}" # the path to the actual file
    object_key = f"{self.cache_dir}/relics/{self.relic_name}/{_key}"
    return (object_key, item_path)

  def _read_state(self):
    if os.path.exists(self._objects_bin_path):
      with open(self._objects_bin_path, "rb") as f:
        self._objects = cloudpickle.load(f)

  def _write_state(self):
    with open(self._objects_bin_path, "wb") as f:
      cloudpickle.dump(self._objects, f)

  """
  Standard set of APIs for put, get, rm, has.
  """

  def put(self, key):
    """The put in case of a local relic only updates the internal map to the key filepath"""
    self._read_state()
    object_key, _ = self.get_id(key)
    self._objects[object_key] = key # reference to existing object
    self._logs.info(f"[{self.workspace_id}/{self.relic_name}] PUT {key}")
    self._write_state()

  def get(self, key: str) -> None:
    self._read_state()
    object_key, _ = self.get_id(key) # internal path
    item_path = self._objects.get(object_key, None)
    if item_path is not None and os.path.exists(item_path):
      self._logs.info(f"[{self.workspace_id}/{self.relic_name}] GET {key}")
    elif (item_path is not None) and (not os.path.exists(item_path)):
      raise Exception("Trying to get missing file")
    else:
      raise Exception("Could not get link, are you sure this file exists?")

  def rm(self, key: str) -> None:
    self._read_state()
    object_key, _ = self.get_id(key) # internal path
    item_path = self._objects.get(object_key, None)
    if item_path is not None and os.path.exists(item_path):
      os.remove(item_path)
      del self._objects[object_key]
      self._logs.warning(f"[{self.workspace_id}/{self.relic_name}] DEL {key}")
      self._write_state()
    elif (item_path is not None) and (not os.path.exists(item_path)):
      raise Exception("Trying to get missing file")
    else:
      raise Exception("Could not get link, are you sure this file exists?")

  def has(self, key: str) -> None:
    self._read_state()
    object_key, _ = self.get_id(key) # internal path
    item_path = self._objects.get(object_key, None)
    if item_path is not None and os.path.exists(item_path):
      return True
    return False

  """
  Conviniece for python objects.
  """

  def put_object(self, key: str, value: bytes) -> None:
    """The put in case of a local relic only updates the internal map to the key filepath"""
    self._read_state()
    object_key, item_path = self.get_id(key) # internal path
    with open(item_path, "wb") as f:
      cloudpickle.dump(value, f)
    self._objects[object_key] = item_path # the path is the actual place to store
    self._logs.info(f"[{self.workspace_id}/{self.relic_name}] PUTO {key}")
    self._write_state()

  def get_object(self, key: str) -> bytes:
    self._read_state()
    object_key, _ = self.get_id(key) # internal path
    item_path = self._objects.get(object_key, None)
    if item_path is not None and os.path.exists(item_path):
      self._logs.info(f"[{self.workspace_id}/{self.relic_name}] GETO {key}")
      with open(item_path, "rb") as f:
        out = cloudpickle.load(f)
      return out
    elif (item_path is not None) and (not os.path.exists(item_path)):
      raise Exception("Trying to get missing file")
    else:
      raise Exception("Could not get link, are you sure this file exists?")

  """
  Some APIs are more on the level of the relic itself.
  """

  def delete(self):
    """Deletes your relic, which in this case means it removes all the entries at that prefix"""
    self._read_state()
    self._logs.warning(f"[{self.workspace_id}/{self.relic_name}] DELR {self.relic_name}")
    object_key, _ = self.get_id() # internal path to relic
    object_relic_key = "/".join(object_key.split("/")[:-1])
    done = False
    for key, item_path in self._objects.items():
      if key.startswith(object_relic_key):
        self.rm(item_path)
        done = True
    if not done:
      raise Exception("Could not delete relic, are you sure it exists?")
    self._write_state()

  def list_files(self, path: str = "") -> List[str]:
    """List all the files in the relic at path"""
    path = path.strip("/")
    object_key, _ = self.get_id(path) # internal path to relic
    object_relic_key = "/".join(object_key.split("/")[:-1])

    # keep only those files that either have the same name or only till one depth ('/')
    all_f = [k for k in self.objects.keys() if k.startswith(object_relic_key)]
    files = set()
    total_slashes = len(object_relic_key.split("/"))
    for f in all_f:
      sections = f.split("/")
      sections = sections[:total_slashes+1]
      files.add("/".join(sections))
    return list(files)



"""This file contains the code for observability in the networking layer."""

from typing import Tuple
import time

try:
  import starlette
  from starlette.requests import Request
  from starlette.responses import Response
  from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
  from starlette.routing import Match
  from starlette.status import HTTP_500_INTERNAL_SERVER_ERROR
  from starlette.types import ASGIApp
except ImportError:
  starlette = None


class GenericMetric:
  # this is like the prometheus client's Counter, Gauge, Histogram, etc. but instead of client
  # worrying about the types, here we are saying that it's the servers headache to worry about
  def __init__(self, name: str):
    pass


class AsgiLmaoMiddleware(BaseHTTPMiddleware):
  def __init__(self, app: ASGIApp, filter_unhandled_paths: bool = False) -> None:
    if starlette is None:
      raise ValueError("Starlette is not installed. pip install -U nbox\[serving\]")
    super().__init__(app)
    self.filter_unhandled_paths = filter_unhandled_paths

  async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
    method = request.method
    path_template, is_handled_path = self.get_path_template(request)

    if self._is_path_filtered(is_handled_path):
      return await call_next(request)

    REQUESTS_IN_PROGRESS.labels(method = method, path_template = path_template).inc()
    REQUESTS.labels(method=method, path_template=path_template).inc()
    before_time = time.perf_counter()
    try:
      response = await call_next(request)
    except BaseException as e:
      status_code = HTTP_500_INTERNAL_SERVER_ERROR
      EXCEPTIONS.labels(method=method, path_template=path_template, exception_type=type(e).__name__).inc()
      raise e from None
    else:
      status_code = response.status_code
      after_time = time.perf_counter()
      REQUESTS_PROCESSING_TIME.labels(method=method, path_template=path_template).observe(after_time - before_time)
    finally:
      RESPONSES.labels(method=method, path_template=path_template, status_code=status_code).inc()
      REQUESTS_IN_PROGRESS.labels(method=method, path_template=path_template).dec()

    return response

  @staticmethod
  def get_path_template(request: Request) -> Tuple[str, bool]:
    for route in request.app.routes:
      match, child_scope = route.matches(request.scope)
      if match == Match.FULL:
        return route.path, True
    return request.url.path, False

  def _is_path_filtered(self, is_handled_path: bool) -> bool:
    return self.filter_unhandled_paths and not is_handled_path





def metrics(request: Request) -> Response:
  if "prometheus_multiproc_dir" in os.environ:
    registry = CollectorRegistry()
    MultiProcessCollector(registry)
  else:
    registry = REGISTRY

  return Response(generate_latest(registry), headers={"Content-Type": CONTENT_TYPE_LATEST})



from starlette_prometheus import metrics, PrometheusMiddleware



"""
``Relic`` is NimbleBox' Object store APIs to connect to your buckets to provide persistant storage beyond instances.
This can support local file storage along with AWS S3, GCP Bucket, Azure Blob Storage, etc.
"""

from nbox.relics.base import BaseStore
from nbox.relics.local import RelicLocal
from nbox.relics.nbx import RelicsNBX


# singluar loader file for all models in nbox

import re
import json
import inspect
from typing import Dict

from .model import Model
from .utils import fetch, isthere

model_key_regex = re.compile(r"^(\w+)(\/[\w\/-]+)?(:*[\w+:]+)?$")

# util functions
def remove_kwargs(model_fn, kwargs):
  """take in the ``model_fn`` and only keep variables from kwargs that can be consumed"""
  # compare variables between the model_fn and kwargs if they are different then remove it with warning
  arg_spec = inspect.getfullargspec(model_fn)
  if kwargs and arg_spec.varkw != None:
    diff = set(kwargs.keys()) - set(arg_spec.args)
    for d in list(diff):
      kwargs.pop(d)
  return kwargs


# --- model loader functions: add your things here --- #
# This is part of experimental API and can be moved to a different place in the future.
# load any nbox.Model object through the pretrained models index, probably migrate them
# to framework and get autogenerated code.

@isthere("efficientnet_pytorch", soft = False)
def _efficientnet_pytorch_models(pretrained=False, **kwargs):
  """`Efficientnet <https://github.com/lukemelas/EfficientNet-PyTorch>`_ package"""
  import efficientnet_pytorch

  if pretrained:
    model_fn = efficientnet_pytorch.EfficientNet.from_pretrained
  else:
    model_fn = efficientnet_pytorch.EfficientNet.from_name

  kwargs = remove_kwargs(model_fn, kwargs)
  return model_fn(**kwargs)


@isthere("torchvision", soft = False)
def _torchvision_models(model, pretrained=False, **kwargs) -> Dict:
  """`torchvision <https://pytorch.org/vision/stable/models.html`_ package"""
  # tv_mr = torchvision models registry
  # this is a json object that maps all the models to their respective methods from torchvision
  # the trick to loading is to eval the method string
  tv_mr = json.loads(fetch("https://raw.githubusercontent.com/NimbleBoxAI/nbox/master/assets/pt_models.json").decode("utf-8"))
  model_fn = tv_mr.get(model, None)
  if model_fn == None:
    raise IndexError(f"Model: {model} not found in torchvision")

  # torchvision is imported here, if it is imported in the outer method, eval fails
  # reason: unknown
  import torchvision

  model_fn = eval(model_fn)
  kwargs = remove_kwargs(model_fn, kwargs)
  model = model_fn(pretrained=pretrained, **kwargs)
  return model


@isthere("transformers", soft = False)
def _transformers_models(model, model_instr, **kwargs):
  """The legendary `transformers <https://huggingface.co/docs/transformers>`_ package"""
  import transformers

  _auto_loaders = {x: getattr(transformers, x) for x in dir(transformers) if x[:4] == "Auto" and x != "AutoConfig"}

  model_instr = model_instr.split("::")
  if len(model_instr) == 1:
    auto_model_type = model_instr[0]
  else:
    # if the task is given, validate that as well
    auto_model_type, task = model_instr
    assert task in [
      "generation",
      "masked_lm",
    ], "For now only the following are supported: `generation`, `masked_lm`"

  # initliase the model and tokenizer object
  tokenizer = transformers.AutoTokenizer.from_pretrained(model, **kwargs)

  # All the tokenizers must contain the pad_token, by default we set it to eos_token
  if not tokenizer.pad_token:
    tokenizer.add_special_tokens({"pad_token": tokenizer.eos_token})

  # tokenizer.pad_token = tokenizer.eos_token if not tokenizer.pad_token_id else tokenizer.pad_token
  model = _auto_loaders[auto_model_type].from_pretrained(model, **kwargs)

  # return the model and tokenizer
  return model, {"tokenizer": tokenizer}


### ----- pretrained models master index
# add code based on conditionals, best way is to only include those that
# have proper model building code like transformers, torchvision, etc.

PRETRAINED_MODELS = {
  "efficientnet_pytorch": _efficientnet_pytorch_models,
  "torchvision": _torchvision_models,
  "transformers": _transformers_models,
}


# ---- load function has to manage everything and return Model object properly initialised


def load(model, pre = None, post = None, verbose=False, **loader_kwargs) -> Model:
  """This function loads the nbox.Model object from the pretrained models index.

  Args:

    model (str): the index of the model to load, you can check with ``nbox.load.PRETRAINED_MODELS``
    pre (str): the preprocessing to apply to the model, it should be callable and should return a dict
    post (str): the postprocessing to apply to the model, it should be callable and should return anything
      you want
    verbose (bool): if True, print out the model name and the model parameters
    loader_kwargs (dict, optional): keyword arguments to be passed to the loader function.

  Returns:
    nbox.Model: the loaded model object
  """
  # the input key can also contain instructions on how to run a particular models and so
  model_key_parts = re.findall(model_key_regex, model)
  if not model_key_parts:
    raise ValueError(f"Key: {model} incorrect, please check!")

  # this key is valid, now get it's components
  src, src_key, model_instr = model_key_parts[0]
  src_key = src_key.strip("/") # remove leading and trailing slashes
  model_instr = model_instr.replace(":", "") # remove the :

  model_fn = PRETRAINED_MODELS.get(src, None)
  if model_fn is None:
    raise IndexError(f"Model: {src} not found")

  # now just load the underlying graph and the model and off you go
  model = model_fn(model=src_key, model_instr=model_instr, **loader_kwargs)
  out = Model(model, pre, post, verbose=verbose)
  return out



"""
With NimbleBox you can run cluster wide workloads from anywhere. This requires capabilities around distributed computing,
process managements. The code here is tested along with ``nbox.Relic`` to perform distributed local processing.
"""

import os
import inspect
import subprocess
from time import sleep

from threading import Thread
from typing import Union

import nbox.utils as U
from nbox import RelicsNBX
from nbox.auth import secret
from nbox.relics.local import RelicLocal
from nbox import Operator, logger
from nbox.operator import Resource
from nbox.utils import log_traceback
from nbox.nbxlib.tracer import Tracer
from nbox.hyperloop.job_pb2 import Job
from nbox import Operator, nbox_grpc_stub
from nbox.messages import rpc, get_current_timestamp
from nbox.hyperloop.nbox_ws_pb2 import UpdateRunRequest
from nbox.nbxlib.serving import serve_operator

PROC_FOLDER = "./nbx_autogen_proc"


# Manager
class RootRunner(Operator):
  def __init__(
    self,
    op: Operator,
    dist = False,
    relic_type = "nbx",
    verbose: str = False,
    *,
    _unittest: bool = False
  ):
    """This is the Operator which runs on the main Job pod and stores all the values. Think of this like the
    Master node in a distributed system, which contains all the information on the processing."""
    super().__init__()
    self.op = op
    self.dist = dist
    self.verbose = verbose
    self._unittest = _unittest

    _relic_cls = {"local": RelicLocal, "nbx": RelicsNBX}.get(relic_type, None)
    if _relic_cls is None:
      raise ValueError(f"Cannot find relic type: '{relic_type}'")

    self.relic = _relic_cls(relic_name = "hustle_bustle", workspace_id = "wnja9glc", create = True)
    self.cache_status = {}

  def start_status(self):
    def worker():
      # function to keep on updating the cache and seeing the progress
      while True:
        available = 0
        for key in self.keys:
          if self.relic.has(key):
            available += 1
        if self.verbose:
          logger.info(f"Available: [{available}/{len(self.keys)}]")
        sleep(0.69)

    # start the status thread
    self.cache_man = Thread(target = worker, daemon=True)
    self.cache_man.start()

  def prepare_children(self):
    self.op_agents = {}
    self.keys = set()
    for op_name, res in self.op._op_to_resource_map.items():
      if res != None:
        # replace the Opeartors in the graph with AgentsStubs
        op = getattr(self.op, op_name)
        agent_op = AgentOpRootStub(op, op_name, res, self.relic)
        setattr(self.op, op_name, agent_op)
        self.op_agents[op_name] = agent_op
        self.keys.add(op_name + "_input")
        self.keys.add(op_name + "_output")

  def clear_relic(self):
    for op_name in self.op_agents:
      for _x in ["_input", "_output"]:
        key = op_name + _x
        if self.relic.has(key):
          self.relic.rm(key)

  def forward(self, *args, **kwargs):
    if self.dist:
      self.prepare_children()
      if self._unittest:
        self.clear_relic()
      self.start_status()

    return self.op(*args, **kwargs)


class AgentOpRootStub(Operator):
  def __init__(self, op: Operator, op_name:str , res: Resource, relic: Union[RelicLocal, RelicsNBX]):
    """This is the root stub for the remote job. In the current implementation it will spin up a new process
    and manage the connections and things for it.
    
    Hmmm... One day maybe [run go in python](https://medium.com/analytics-vidhya/running-go-code-from-python-a65b3ae34a2d)
    
    Think of this like kubelet or raylet"""
    super().__init__()
    self.op = op
    self.op_name = op_name
    self.relic = relic
    logger.info(f"Initialising AgentOpRootStub: {self.op_name}")

  def get_keys(self):
    input_key = self.op_name +  "_input"
    output_key = self.op_name + "_output"
    return input_key, output_key

  def forward(self, *args, **kwargs):
    # take this op and deploy this to a remote machine where it will start
    # and wait for the remote to finish

    # print(self.relic._objects)
    input_key, output_key = self.get_keys()
    if self.relic.has(output_key):
      self.relic.rm(output_key)
    logger.info(f"{input_key} / {output_key}")

    if not os.path.exists(PROC_FOLDER):
      os.mkdir(PROC_FOLDER)

    fpath = os.path.join(PROC_FOLDER, "auto_" + self.op_name + ".py")
    logger.info(f"Creating file: {fpath}")
    with open(fpath, "w") as f:
      f.write(f'''# Auto generated
import os

os.environ["NBOX_LOG_LEVEL"] = "debug"

from nbox import operator
from nbox.relics import RelicLocal, RelicsNBX
from nbox.utils import load_module_from_path

@operator()
def run():
  # load the operator
  op_cls = load_module_from_path("{self.op.__class__.__qualname__}", "{inspect.getfile(self.op.__class__)}")
  op = op_cls()

  # define the keys
  op_name = "{self.op_name}"
  input_key = "{input_key}"
  output_key = "{output_key}"
  
  # define relic and get things
  relic = RelicsNBX(relic_name = "hustle_bustle", workspace_id = "wnja9glc", create = True)
  args, kwargs = relic.get_object(input_key)
  
  out = op(*args, **kwargs)
  if out == None:
    out = "NBX_NULL"

  # print("Putting", output_key, out)
  relic.put_object(output_key, out)

if __name__ == "__main__":
  # if required user can update the status of the operator
  # run.thaw()
  run()
''')

    # put the data in the object store
    self.relic.put_object(input_key, (list(args), kwargs))

    # run the file as a subprocess: pseudo parallel
    try:
      out = subprocess.run(["python3", fpath], universal_newlines=True)
      if out.returncode != 0:
        raise Exception(f"Process returned non zero code: {out.returncode}")
    except KeyboardInterrupt:
      logger.info("KeyboardInterrupt, closing connections")
    except Exception as e:
      logger.error(f"Error in process: '{self.op_name}', check above for details")
      log_traceback()
      raise e

    # get the results from the object store
    out = self.relic.get_object(output_key)
    while out == None:
      sleep(1)
      out = self.relic.get_object(output_key)
    return out


class NBXLet(Operator):
  def __init__(self, op: Operator):
    """The Operator that runs the things on any pod on the NimbleBox Jobs + Deploy platform.
    Name mimics the kubelet, dockerlet, raylet, etc"""
    super().__init__()
    self.op = op

  def run(self):
    """Run this as a batch process"""
    tracer = Tracer()
    secret.put("username", tracer.job_proto.auth_info.username)

    self.op.propagate(_tracer = tracer)
    if hasattr(self.op._tracer, "job_proto"):
      self.op.thaw(self.op._tracer.job_proto)

    workspace_id = tracer.job_proto.auth_info.workspace_id
    job_id = tracer.job_id
    status = Job.Status.ERROR

    try:
      # now for some jobs there might be a relic Object so we can check if that exists, it will always
      # be present in the dot_deploy_cache folder and will be in the {job_id} folder
      relic = RelicsNBX("cache", workspace_id, create = True)

      # check if there is a specific relic for this job
      run_tag = os.getenv("NBOX_RUN_METADATA", None)
      _in = f"{job_id}/args_kwargs"
      if run_tag:
        _in += f"_{run_tag}"
      if relic.has(_in):
        (args, kwargs) = relic.get_object(_in)
      else:
        args, kwargs = (), {}
      out = self.op(*args, **kwargs)
      _out = f"{job_id}/return"
      if run_tag:
        _out += f"_{run_tag}"
      relic.put_object(_out, out)
      status = Job.Status.COMPLETED
    except Exception as e:
      U.log_traceback()
    finally:
      logger.info(f"Job {job_id} completed with status {status}")
      if hasattr(tracer, "job_proto"):
        self.op._tracer.job_proto.status = status
        rpc(
          nbox_grpc_stub.UpdateRun, UpdateRunRequest(
            token = tracer.run_id, job = tracer.job_proto, updated_at = get_current_timestamp()
          ), "Failed to end job!"
        )
      U._exit_program()

  def serve(self, host: str = "0.0.0.0", port: int = 8000, *, model_name: str = None):
    """Run a serving API endpoint"""
    try:
      serve_operator(self.op, host = host, port = port, model_name = model_name)
    except Exception as e:
      U.log_traceback()
      logger.error(f"Failed to serve operator: {e}")
      U._exit_program()



"""
``nbox.Job`` and ``nbox.Serve`` are wrappers to the NBX-Jobs and NBX-Deploy APIs and contains staticmethods for convinience from the CLI.

Notes
-----

* ``datetime.now(timezone.utc)`` is incorrect, use `this <https://blog.ganssle.io/articles/2019/11/utcnow.html>`_ method.
"""

import os
import sys
import jinja2
import tabulate
from functools import partial
from datetime import datetime, timedelta, timezone
from google.protobuf.timestamp_pb2 import Timestamp
from google.protobuf.field_mask_pb2 import FieldMask

import nbox.utils as U
from nbox.auth import secret, ConfigString
from nbox.utils import logger
from nbox.version import __version__
from nbox.messages import rpc, streaming_rpc
from nbox.hyperloop.nbox_ws_pb2 import JobInfo
from nbox.init import nbox_grpc_stub, nbox_ws_v1

from nbox.hyperloop.job_pb2 import NBXAuthInfo, Job as JobProto, Resource
from nbox.hyperloop.nbox_ws_pb2 import ListJobsRequest, JobLogsRequest, ListJobsResponse, UpdateJobRequest


################################################################################
"""
# Common Functions

These functions are common to both NBX-Jobs and NBX-Deploy.
"""
################################################################################

def _repl_schedule(return_proto: bool = False):
  """Create a schedule from the user input.

  Args:
    return_proto (bool, optional): If `True` returns proto otherwise returns cron string.
  """
  logger.info("Calendar Instructions (all time is in UTC Timezone):")
  logger.info("            What you want: Code")
  logger.info("\"   every 4:30 at friday\": Schedule(4, 30, ['fri'])")
  logger.info("\"every 12:00 on weekends\": Schedule(12, 0, ['sat', 'sun'])")
  logger.info("\"         every 10 hours\": Schedule(10)")
  logger.info("\"      every 420 minutes\": Schedule(minute = 420)")
  logger.info("> Enter the calender instruction: ")
  cron_instr = input("> ").strip()
  schedule_proto = None
  try:
    schedule_proto: Schedule = eval(cron_instr, {'Schedule': Schedule})
  except Exception as e:
    logger.error(f"Invalid calender instruction: {e}")
    sys.exit(1)
  if not return_proto:
    return cron_instr
  return schedule_proto

def new(folder_name):
  """This creates a single folder that can be used with both NBX-Jobs and NBX-Deploy.

  Args:
    folder_name (str): The name of the job folder
  """
  folder_name = str(folder_name)
  if folder_name == "":
    raise ValueError("Folder name can not be empty")
  if os.path.exists(folder_name):
    raise ValueError(f"Project {folder_name} already exists")
  os.makedirs(folder_name)

  # get a timestamp like this: Monday W34 [UTC 12 April, 2022 - 12:00:00]
  _ct = datetime.now(timezone.utc)
  _day = ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"][_ct.weekday()]
  created_time = f"{_day} W{_ct.isocalendar()[1]} [ UTC {_ct.strftime('%d %b, %Y - %H:%M:%S')} ]"

  # we no longer need to ask for the deployment ID / Job ID since deployment is now concern of
  # nbx [jobs/serve] upload ...

  # create the necessary files that can be used
  os.chdir(folder_name)
  with open(".nboxignore", "w") as f:
    f.write('''# This is where you list the files that you don't want to upload with your code

__pycache__/
    ''')
  
  with open("requirements.txt", "w") as f:
    f.write(f"""# This is where you list the third-party dependencies of your pipeline

# Examples:
# pandas

# always be on lookout for optimisations, ex:
# save 85% download if using CPU torch add `-f https://download.pytorch.org/whl/torch_stable.html`

nbox[serving]=={__version__} # do not change this
""")

  assets = U.join(U.folder(__file__), "assets")
  path = U.join(assets, "user.jinja")
  with open(path, "r") as f, open("nbx_user.py", "w") as f2:
    f2.write(jinja2.Template(f.read()).render({
      "created_time": created_time,
      "nbx_username": secret.get("username"),
    }))

  logger.info(f"Created folder: {folder_name}")

def upload_job_folder(method: str, init_folder: str, id_or_name: str, *, workspace_id: str = ""):
  """Upload the code for a job to the NBX-Jobs if not present, it will create a new Job.

  Args:
    init_folder (str): folder with all the relevant files
    id_or_name (str): Job ID or name
    workspace_id (str, optional): Workspace ID, `None` for personal workspace. Defaults to None.
  """
  workspace_id = workspace_id or secret.get(ConfigString.workspace_id.value)
  if init_folder not in sys.path:
    sys.path.append(init_folder)
  from nbx_user import get_op, get_resource, get_schedule

  from nbox import Operator
  from nbox.network import deploy_job, deploy_serving
  operator: Operator = get_op(method == "serving")

  if method == "jobs":
    out: Job = deploy_job(
      init_folder = init_folder,
      job_id_or_name = id_or_name,
      dag = operator._get_dag(),
      workspace_id = workspace_id,
      schedule = get_schedule(),
      resource = get_resource(),
      _unittest = False
    )
    return out
  elif method == "serving":
    out: Serve = deploy_serving(
      init_folder = init_folder,
      deployment_id_or_name = id_or_name,
      workspace_id = workspace_id,
      resource = get_resource(),
      wait_for_deployment = False,
      _unittest = False
    )
  
  return out

################################################################################
"""
# NimbleBox.ai Serving

This is the proposed interface for the NimbleBox.ai Serving API. We want to keep
the highest levels of consistency with the NBX-Jobs API.
"""
################################################################################

def _get_deployment_data(id_or_name, *, workspace_id: str = ""):
  # filter and get "id" and "name"
  workspace_id = workspace_id or secret.get(ConfigString.workspace_id.value)
  stub_all_depl = nbox_ws_v1.workspace.u(workspace_id).deployments
  all_deployments = stub_all_depl()
  models = list(filter(lambda x: x["deployment_id"] == id_or_name or x["deployment_name"] == id_or_name, all_deployments))
  if len(models) == 0:
    logger.info(f"No Job found with ID or name: {id_or_name}, will create a new one")
    serving_name = id_or_name
    serving_id = None
  elif len(models) > 1:
    raise ValueError(f"Multiple models found for '{id_or_name}', try passing ID")
  else:
    logger.info(f"Found deployment with ID or name: {id_or_name}, will update it")
    data = models[0]
    serving_name = data["deployment_name"]
    serving_id = data["deployment_id"]
  return serving_id, serving_name


def print_serving_list(sort: str = "created_on", *, workspace_id: str = ""):
  def _get_time(t):
    return datetime.fromtimestamp(int(float(t))).strftime("%Y-%m-%d %H:%M:%S")

  workspace_id = workspace_id or secret.get(ConfigString.workspace_id.value)
  stub_all_depl = nbox_ws_v1.workspace.u(workspace_id).deployments
  all_deployments = stub_all_depl()
  sorted_depls = sorted(all_deployments, key = lambda x: x[sort], reverse = sort == "created_on")
  headers = ["created_on", "id", "name", "pinned_id", "pinned_name", "pinned_last_updated"]
  all_depls = []
  for depl in sorted_depls:
    _depl = (_get_time(depl["created_on"]), depl["deployment_id"], depl["deployment_name"],)
    pinned = depl["pinned_model"]
    if not pinned:
      _depl += (None, None,)
    else:
      _depl += (pinned["id"], pinned["name"], _get_time(pinned["last_updated"]),)
    all_depls.append(_depl)

  for l in tabulate.tabulate(all_depls, headers).splitlines():
    logger.info(l)

  with open("requirements.txt", "w") as f:
    f.write(f"nbox=={__version__}")

# def serving_forward(id_or_name: str, token: str, workspace_id: str = "", **kwargs):
#   workspace_id = workspace_id or secret.get(ConfigString.workspace_id.value)
#   from nbox.operator import Operator
#   op = Operator.from_serving(id_or_name, token, workspace_id)
#   out = op(**kwargs)
#   logger.info(out)

class Serve:
  new = staticmethod(new) # create a new folder, alias but `jobs new` should be used
  status = staticmethod(print_serving_list)
  upload = staticmethod(partial(upload_job_folder, "serving"))

  def __init__(self, id, model_id: str = None, *, workspace_id: str = "") -> None:
    """Python wrapper for NBX-Serving gRPC API

    Args:
      id (str): Deployment ID
      workspace_id (str, optional): If None personal workspace is used. Defaults to None.
    """
    self.id = id
    self.model_id = model_id
    self.workspace_id = workspace_id or secret.get(ConfigString.workspace_id.value)
    if workspace_id is None:
      raise DeprecationWarning("Personal workspace does not support serving")
    else:
      serving_id, serving_name = _get_deployment_data(self.id, workspace_id = self.workspace_id)
    self.serving_id = serving_id
    self.serving_name = serving_name
    self.ws_stub = nbox_ws_v1.workspace.u(workspace_id).deployments
    
  # def change_behaviour(self, new_behaviour: 'Behaviour' = None):
  #   raise NotImplementedError("Not implemented yet")

  def __repr__(self) -> str:
    x = f"nbox.Serve('{self.id}', '{self.workspace_id}'"
    if self.model_id is not None:
      x += f", model_id = '{self.model_id}'"
    x += ")"
    return x

   

################################################################################
"""
# NimbleBox.ai Jobs

This is the actual job object that users can manipulate. It is a shallow class
around the NBX-Jobs gRPC API.
"""
################################################################################

def _get_job_data(id_or_name, *, workspace_id: str = ""):
  # get stub
  workspace_id = workspace_id or secret.get(ConfigString.workspace_id.value)
  if workspace_id == None:
    stub_all_jobs = nbox_ws_v1.user.jobs
  else:
    stub_all_jobs = nbox_ws_v1.workspace.u(workspace_id).jobs

  # filter and get "id" and "name"
  all_jobs = stub_all_jobs()
  jobs = list(filter(lambda x: x["job_id"] == id_or_name or x["name"] == id_or_name, all_jobs))
  if len(jobs) == 0:
    job_name =  id_or_name
    job_id = None
    logger.info(f"No Job found with ID or name: {job_name}")
  elif len(jobs) > 1:
    raise ValueError(f"Multiple jobs found for '{id_or_name}', try passing ID")
  else:
    data = jobs[0]
    job_name = data["name"]
    job_id = data["job_id"]
    logger.info(f"Found job with ID '{job_id}' and name '{job_name}'")
  
  return job_id, job_name


def get_job_list(sort: str = "name", *, workspace_id: str = ""):
  """Get list of jobs, optionally in a workspace"""
  workspace_id = workspace_id or secret.get(ConfigString.workspace_id.value)

  def _get_time(t):
    return datetime.fromtimestamp(int(float(t))).strftime("%Y-%m-%d %H:%M:%S")

  auth_info = NBXAuthInfo(workspace_id = workspace_id)
  out: ListJobsResponse = rpc(
    nbox_grpc_stub.ListJobs,
    ListJobsRequest(auth_info = auth_info),
    "Could not get job list",
  )

  if len(out.Jobs) == 0:
    logger.info("No jobs found")
    sys.exit(0)

  headers = ['created_at', 'id', 'name', 'schedule', 'status']
  try:
    sorted_jobs = sorted(out.Jobs, key = lambda x: getattr(x, sort))
  except:
    logger.error(f"Cannot sort on key: {sort}")
  data = []
  for j in sorted_jobs:
    _row = []
    for x in headers:
      if x == "status":
        _row.append(JobProto.Status.keys()[getattr(j, x)])
        continue
      elif x == "created_at":
        _row.append(_get_time(j.created_at.seconds))
      elif x == "schedule":
        _row.append(j.schedule.cron)
      else:
        _row.append(getattr(j, x))
    data.append(_row)
  for l in tabulate.tabulate(data, headers).splitlines():
    logger.info(l)

class Job:
  new = staticmethod(new)
  status = staticmethod(get_job_list)
  upload = staticmethod(partial(upload_job_folder, "jobs"))

  def __init__(self, id, *, workspace_id: str = ""):
    """Python wrapper for NBX-Jobs gRPC API

    Args:
        id (str): job ID
        workspace_id (str, optional): If None personal workspace is used. Defaults to None.
    """
    workspace_id = workspace_id or secret.get(ConfigString.workspace_id.value)
    self.id, self.name = _get_job_data(id, workspace_id = workspace_id)
    self.workspace_id = workspace_id
    self.job_proto = JobProto(id = self.id, auth_info = NBXAuthInfo(workspace_id = workspace_id))
    self.job_info = JobInfo(job = self.job_proto)

    self.run_stub = None
    self.runs = []

    # sometimes a Job can be a placeholder and not have any data against it, thus it won't make
    # sense to try to get the data. This causes RPC error 
    if self.id is not None:
      self.refresh()
      self.get_runs() # load the runs as well

  @property
  def exists(self):
    return self.id is not None

  def change_schedule(self, new_schedule: 'Schedule' = None):
    """Change schedule this job"""
    logger.debug(f"Updating job '{self.job_proto.id}'")
    if new_schedule == None:
      new_schedule = _repl_schedule(True) # get the information from REPL
    self.job_proto.schedule.MergeFrom(new_schedule.get_message())
    rpc(
      nbox_grpc_stub.UpdateJob,
      UpdateJobRequest(job=self.job_proto, update_mask=FieldMask(paths=["schedule"])),
      "Could not update job schedule",
      raise_on_error = True
    )
    logger.debug(f"Updated job '{self.job_proto.id}'")
    self.refresh()

  def __repr__(self) -> str:
    x = f"nbox.Job('{self.job_proto.id}', '{self.job_proto.auth_info.workspace_id}'): {self.status}"
    if self.job_proto.schedule.ByteSize != None:
      x += f" {self.job_proto.schedule}"
    else:
      x += " (no schedule)"
    return x

  def logs(self, f = sys.stdout):
    """Stream logs of the job, ``f`` can be anything has a ``.write/.flush`` methods"""
    logger.debug(f"Streaming logs of job '{self.job_proto.id}'")
    for job_log in streaming_rpc(
      nbox_grpc_stub.GetJobLogs,
      JobLogsRequest(job = JobInfo(job = self.job_proto)),
      f"Could not get logs of job {self.job_proto.id}, is your job complete?",
      True
    ):
      for log in job_log.log:
        f.write(log + "\n")
        f.flush()

  def delete(self):
    """Delete this job"""
    logger.info(f"Deleting job '{self.job_proto.id}'")
    rpc(nbox_grpc_stub.DeleteJob, JobInfo(job = self.job_proto,), "Could not delete job")
    logger.info(f"Deleted job '{self.job_proto.id}'")
    self.refresh()

  def refresh(self):
    """Refresh Job statistics"""
    logger.debug(f"Updating job '{self.job_proto.id}'")
    if self.id == None:
      self.id, self.name = _get_job_data(self.id, workspace_id = self.workspace_id)
    if self.id == None:
      return
      
    self.job_proto: JobProto = rpc(
      nbox_grpc_stub.GetJob, JobInfo(job = self.job_proto), f"Could not get job {self.job_proto.id}"
    )
    self.job_proto.auth_info.CopyFrom(NBXAuthInfo(workspace_id = self.workspace_id))
    self.job_info.CopyFrom(JobInfo(job = self.job_proto))
    logger.debug(f"Updated job '{self.job_proto.id}'")

    self.status = self.job_proto.Status.keys()[self.job_proto.status]

  def trigger(self, tag: str = ""):
    """Manually triger this job"""
    logger.debug(f"Triggering job '{self.job_proto.id}'")
    if tag:
      self.job_proto.feature_gates.update({"SetRunMetadata": tag})
    rpc(nbox_grpc_stub.TriggerJob, JobInfo(job = self.job_proto), f"Could not trigger job '{self.job_proto.id}'")
    logger.info(f"Triggered job '{self.job_proto.id}'")
    self.refresh()

  def pause(self):
    """Pause the execution of this job.
    
    WARNING: This will "cancel" all the scheduled runs, if present""" 
    logger.info(f"Pausing job '{self.job_proto.id}'")
    job: JobProto = self.job_proto
    job.status = JobProto.Status.PAUSED
    rpc(nbox_grpc_stub.UpdateJob, UpdateJobRequest(job=job, update_mask=FieldMask(paths=["status", "paused"])), f"Could not pause job {self.job_proto.id}", True)
    logger.debug(f"Paused job '{self.job_proto.id}'")
    self.refresh()
  
  def resume(self):
    """Resume the Job with the current schedule, if provided else simlpy sets status as ACTIVE"""
    logger.info(f"Resuming job '{self.job_proto.id}'")
    job: JobProto = self.job_proto
    job.status = JobProto.Status.SCHEDULED
    rpc(nbox_grpc_stub.UpdateJob, UpdateJobRequest(job=job, update_mask=FieldMask(paths=["status", "paused"])), f"Could not resume job {self.job_proto.id}", True)
    logger.debug(f"Resumed job '{self.job_proto.id}'")
    self.refresh()

  def _get_runs(self, page = -1, limit = 10):
    self.run_stub = nbox_ws_v1.workspace.u(self.workspace_id).job.u(self.id).runs
    runs = self.run_stub(limit = limit, page = page)["runs_list"]
    return runs

  def get_runs(self, page = -1, sort = "s_no", limit = 10):
    runs = self._get_runs(page, limit)
    sorted_runs = sorted(runs, key = lambda x: x[sort])
    return sorted_runs

  def display_runs(self, sort: str = "created_at", page: int = -1, limit = 10):
    headers = ["s_no", "created_at", "run_id", "status"]

    def _display_runs(runs):
      data = []
      for run in runs:
        data.append((run["s_no"], run["created_at"], run["run_id"], run["status"]))
      for l in tabulate.tabulate(data, headers).splitlines():
        logger.info(l)

    runs = self.get_runs(sort = sort, page = page, limit = limit) # time should be reverse
    _display_runs(runs)
    if page == -1:
      page = 1
    y = input(f">> Print {limit} more runs? (y/n): ")
    done = y != "y"
    while not done:
      _display_runs(self.get_runs(page = page + 1, sort = sort, limit = limit))
      page += 1
      y = input(f">> Print {limit} more runs? (y/n): ")
      done = y != "y"

  def last_n_runs(self, n: int = 10):
    all_items = []
    _page = 1
    out = self._get_runs(_page, 20)
    all_items.extend(out)

    while len(all_items) < n:
      _page += 1
      out = self._get_runs(_page, 20)
      if not len(out):
        break
      all_items.extend(out)

    if n == 1:
      return all_items[0]
    return all_items

class Schedule:
  def __init__(
    self,
    hour: int  = None,
    minute: int = None,
    days: list = [],
    months: list = [],
    starts: datetime = None,
    ends: datetime = None,
  ):
    """Make scheduling natural.

    Usage:

    .. code-block:: python

      # 4:20 everyday
      Schedule(4, 0)

      # 4:20 every friday
      Schedule(4, 20, ["fri"])

      # 4:20 every friday from jan to feb
      Schedule(4, 20, ["fri"], ["jan", "feb"])

      # 4:20 everyday starting in 2 days and runs for 3 days
      starts = datetime.now(timezone.utc) + timedelta(days = 2) # NOTE: that time is in UTC
      Schedule(4, 20, starts = starts, ends = starts + timedelta(days = 3))

      # Every 1 hour
      Schedule(1)

      # Every 69 minutes
      Schedule(minute = 69)

    Args:
        hour (int): Hour of the day, if only this value is passed it will run every ``hour``
        minute (int): Minute of the hour, if only this value is passed it will run every ``minute``
        days (str/list, optional): List of days (first three chars) of the week, if not passed it will run every day.
        months (str/list, optional): List of months (first three chars) of the year, if not passed it will run every month.
        starts (datetime, optional): UTC Start time of the schedule, if not passed it will start now.
        ends (datetime, optional): UTC End time of the schedule, if not passed it will end in 7 days.
    """
    self.hour = hour
    self.minute = minute

    self._is_repeating = self.hour or self.minute
    self.mode = None
    if self.hour == None and self.minute == None:
      raise ValueError("Atleast one of hour or minute should be passed")
    elif self.hour != None and self.minute != None:
      assert self.hour in list(range(0, 24)), f"Hour must be in range 0-23, got {self.hour}"
      assert self.minute in list(range(0, 60)), f"Minute must be in range 0-59, got {self.minute}"
    elif self.hour != None:
      assert self.hour in list(range(0, 24)), f"Hour must be in range 0-23, got {self.hour}"
      self.mode = "every"
      self.minute = datetime.now(timezone.utc).strftime("%m") # run every this minute past this hour
      self.hour = f"*/{self.hour}"
    elif self.minute != None:
      self.hour = self.minute // 60
      assert self.hour in list(range(0, 24)), f"Hour must be in range 0-23, got {self.hour}"
      self.minute = f"*/{self.minute % 60}"
      self.hour = f"*/{self.hour}" if self.hour > 0 else "*"
      self.mode = "every"

    _days = {k:str(i) for i,k in enumerate(["sun","mon","tue","wed","thu","fri","sat"])}
    _months = {k:str(i+1) for i,k in enumerate(["jan","feb","mar","apr","may","jun","jul","aug","sep","oct","nov","dec"])}

    if isinstance(days, str):
      days = [days]
    if isinstance(months, str):
      months = [months]

    diff = set(days) - set(_days.keys())
    if len(diff):
      raise ValueError(f"Invalid days: {diff}")
    self.days = ",".join([_days[d] for d in days]) if days else "*"

    diff = set(months) - set(_months.keys())
    if len(diff):
      raise ValueError(f"Invalid months: {diff}")
    self.months = ",".join([_months[m] for m in months]) if months else "*"

    self.starts = starts or datetime.now(timezone.utc)
    self.ends = ends or datetime.now(timezone.utc) + timedelta(days = 7)

  @property
  def cron(self):
    """Cron string"""
    if self.mode == "every":
      return f"{self.minute} {self.hour} * * *"
    return f"{self.minute} {self.hour} * {self.months} {self.days}"

  def get_dict(self):
    return {"cron": self.cron, "mode": self.mode, "starts": self.starts, "ends": self.ends}

  def get_message(self) -> JobProto.Schedule:
    """Get the JobProto.Schedule object for this Schedule"""
    _starts = Timestamp(); _starts.GetCurrentTime()
    _ends = Timestamp(); _ends.FromDatetime(self.ends)
    # return JobProto.Schedule(start = _starts, end = _ends, cron = self.cron)
    return JobProto.Schedule(cron = self.cron)

  def __repr__(self):
    return str(self.get_dict())
